from typing import Any, Union, Optional
from collections import OrderedDict
import unittest
import numpy as np
from .backends import ImageBackend, builtin_image_backends

__all__ = ["ImageTestCase"]


class ImageTestCase(unittest.TestCase):
    """Testcase for unit testing with images.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.backends = OrderedDict(())
        for name, backend in builtin_image_backends().items():
            self.add_image_backend(name, backend)

    def add_image_backend(
        self, name: str, backend: ImageBackend, allow_duplicate_type: bool = False
    ) -> None:
        """Adds custom :class:`~pyimagetest.backends.backend.ImageBackend` to
        the available backends.

        Args:
            name: Name of the backend
            backend: Backend
            allow_duplicate_type: If ``True``, no check for duplicate
             :attr:`~pyimagetest.backends.backend.ImageBackend.native_image_type` s
             is performed. Defaults to ``False``.

        Raises:
            RuntimeError: If another :class:`~pyimagetest.backends.backend.ImageBackend`
                with the same
                :attr:`~pyimagetest.backends.backend.ImageBackend.native_image_type`
                already present and ``allow_duplicate_type`` is ``False``.

        .. note::
            If you add an :class:`~pyimagetest.backends.backend.ImageBackend`
            with a duplicate
            :attr:`~pyimagetest.backends.backend.ImageBackend.native_image_type`,
            the automatic backend inference with
            :meth:`~pyimagetest.image_test_case.ImageTestCase.infer_image_backend`
            might not work correctly.
        """
        native_image_types = [
            backend.native_image_type for backend in self.backends.values()
        ]
        if not allow_duplicate_type and backend.native_image_type in native_image_types:
            msg = (
                f"Another backend with native_image_type {backend.native_image_type} "
                f"is already present."
            )
            raise RuntimeError(msg)
        self.backends[name] = backend

    def remove_image_backend(self, name: str) -> None:
        """Removes an :class:`~pyimagetest.backends.backend.ImageBackend` from
        the known backends.

        Args:
            name: Name of the backend to be removed
        """
        del self.backends[name]

    def infer_image_backend(self, image: Any) -> ImageBackend:
        """Infers the corresponding :class:`~pyimagetest.backends.backend.ImageBackend`
        from ``image``.

        Args:
            image: Image with type of any known backend

        Raises:
            RuntimeError: If type of image does not correspond to any known
                image backend
        """
        for backend in self.backends.values():
            if image in backend:
                return backend
        msg = f"No backend with native_image_type {type(image)} is known."
        raise RuntimeError(msg)

    def default_image_file(self) -> str:
        """If overwritten, should return the default image file.

        .. note::
            Override this if you mainly work with a single image. If you do,
            :meth:`~pyimagetest.image_test_case.ImageTestCase.load_image` will
            use the value returned by this method if the ``file`` parameter is
            omitted.

        Raises:
            RuntimeError: If not overridden
        """
        raise RuntimeError

    def default_image_backend(self) -> Union[ImageBackend, str]:
        """If overwritten, should return the default
        :class:`~pyimagetest.backends.backend.ImageBackend` or its name.

        .. note::
            Override this if you mainly work with a single backend. If you do,
            :meth:`~pyimagetest.image_test_case.ImageTestCase.load_image` will
            use the value returned by this method if the ``backend`` parameter
            is omitted.

        Raises:
            RuntimeError: If not overridden
        """
        raise RuntimeError

    def load_image(
        self,
        file: Optional[str] = None,
        backend: Optional[Union[ImageBackend, str]] = None,
    ) -> Any:
        """Loads the image from ``file`` with ``backend``. If no ``file`` or
        ``backend`` is given the default values from
        :meth:`~pyimagetest.image_test_case.ImageTestCase.default_image_file`
        and
        :meth:`~pyimagetest.image_test_case.ImageTestCase.default_image_backend`
        are used.

        Args:
            backend: Backend or backend name. If ``None``,
                :meth:`~pyimagetest.image_test_case.ImageTestCase.default_image_backend`
                is used. Defaults to ``None``.
            file: Path to image file. If ``None``,
                :meth:`~pyimagetest.image_test_case.ImageTestCase.default_image_file`
                is used. Defaults to ``None``.
        """

        def parse_file(file: Optional[str]) -> str:
            if isinstance(file, str):
                return file
            elif file is None:
                try:
                    return self.default_image_file()
                except RuntimeError:
                    msg = (
                        "Override ImageTestCase.default_image_file() to be able "
                        "to call ImageTestCase.load_image() without file parameter."
                    )
                    raise RuntimeError(msg)
            else:
                raise TypeError

        def parse_backend(backend: Optional[Union[ImageBackend, str]]) -> ImageBackend:
            if isinstance(backend, ImageBackend):
                return backend

            if backend is None:
                try:
                    backend = self.default_image_backend()
                except RuntimeError:
                    msg = (
                        "Override ImageTestCase.default_image_backend() to be able "
                        "to call ImageTestCase.load_image() without backend."
                        "parameter"
                    )
                    raise RuntimeError(msg)

            if isinstance(backend, str):
                return self.backends[backend]
            else:
                raise TypeError

        file = parse_file(file)
        backend = parse_backend(backend)
        return backend.import_image(file)

    def assertImagesAlmostEqual(
        self,
        image1: Any,
        image2: Any,
        tolerance: float = 1e-2,
        backend1: Optional[Union[ImageBackend, str]] = None,
        backend2: Optional[Union[ImageBackend, str]] = None,
    ):
        """Image equality assertion.

        Args:
            image1: Image 1
            image2: Image 2
            tolerance: Acceptable `mean absolute error (MAE)
                <https://en.wikipedia.org/wiki/Mean_absolute_error>`_. Defaults
                to ``1e-2``.
            backend1:
                :class:`~pyimagetest.backends.backend.ImageBackend` or its name
                for ``image1``. If ``None``, the backend is inferred from
                ``image1`` with
                :meth:`~pyimagetest.image_test_case.ImageTestCase.infer_image_backend`.
                Defaults to ``None``.
            backend2:
                :class:`~pyimagetest.backends.backend.ImageBackend` or its name
                for ``image2``. If ``None``, the backend is inferred from
                ``image2`` with
                :meth:`~pyimagetest.image_test_case.ImageTestCase.infer_image_backend`.
                Defaults to ``None``.

        Raises:
            AssertionError: If `image1` and `image2` are not equal up to the
                desired tolerance.
        """

        def parse_backend(
            backend: Optional[Union[ImageBackend, str]], image: Any
        ) -> ImageBackend:
            if isinstance(backend, ImageBackend):
                return backend
            elif isinstance(backend, str):
                return self.backends[backend]
            elif backend is None:
                return self.infer_image_backend(image)
            else:
                raise RuntimeError

        backend1 = parse_backend(backend1, image1)
        backend2 = parse_backend(backend2, image2)

        image1 = backend1.export_image(image1)
        image2 = backend2.export_image(image2)

        actual = np.mean(np.abs(image1 - image2))
        desired = 0.0
        np.testing.assert_allclose(actual, desired, atol=tolerance, rtol=0.0)



from os import path
from itertools import combinations
from pyimagetest import ImageTestCase
from pyimagetest import builtin_image_backends
from pyimagetest.backends.builtin import BUILTIN_IMAGE_BACKENDS_META


class Tester(ImageTestCase):
    def default_image_file(self) -> str:
        # The test image was downloaded from
        # http://www.r0k.us/graphics/kodak/kodim15.html
        # and is cleared for unrestricted usage
        here = path.abspath(path.dirname(__file__))
        return path.join(here, "test_image.png")

    def test_backend_availability(self) -> None:
        backend_names = [meta.name for meta in BUILTIN_IMAGE_BACKENDS_META]
        available_backend_names = builtin_image_backends().keys()
        not_available_backend_names = set(backend_names) - set(available_backend_names)
        if not_available_backend_names:
            msg = "The following backends are not available for testing: {}"
            raise RuntimeError(msg.format(", ".join(not_available_backend_names)))

    def test_io(self) -> None:
        for backend1, backend2 in combinations(self.backends.values(), 2):
            image1 = self.load_image(backend=backend1)
            image2 = self.load_image(backend=backend2)
            self.assertImagesAlmostEqual(image1, image2)



from typing import Type
import numpy as np
from PIL import Image
from .backend import ImageBackend


class PILBackend(ImageBackend):
    """:class:`~pyimagetest.backends.backend.ImageBackend` for the
    `Pillow package <https://python-pillow.org/>`_ .
    """

    @property
    def native_image_type(self) -> Type[Image.Image]:
        return Image.Image

    def import_image(self, file: str) -> Image.Image:
        return Image.open(file)

    def export_image(self, image: Image.Image) -> np.ndarray:
        mode = image.mode
        image = np.asarray(image, dtype=np.float32)
        if mode in ("L", "RGB"):
            image /= 255.0
        if mode in ("1", "L"):
            image = np.expand_dims(image, 2)
        return image



__all__ = [
    "__name__",
    "__description__",
    "__version__",
    "__url__",
    "__license__",
    "__author__",
    "__author_email__",
]

__name__ = "pyimagetest"
__description__ = "Utilities for unit testing with images"
__version__ = "0.2.post0"
__url__ = "https://github.com/pmeier/pyimagetest"
__license__ = "BSD-3"
__author__ = "Philip Meier"
__author_email__ = "github.pmeier@posteo.de"



from abc import ABC, abstractmethod
from typing import Any, Type
import numpy as np

__all__ = ["ImageBackend"]


class ImageBackend(ABC):
    """ABC for image backends. Each subclass has to implement the
    :attr:`~pyimagetest.backends.backend.ImageBackend.native_image_type` as
    well as the basic I/O methods
    :meth:`~pyimagetest.backends.backend.ImageBackend.import_image`
    and :meth:`~pyimagetest.backends.backend.ImageBackend.export_image`.
    """

    @property
    @abstractmethod
    def native_image_type(self) -> Type[Any]:
        """Native image type of the backend. This is used to infer the
        :class:`~pyimagetest.backends.backend.ImageBackend` from a given image.
        """
        pass

    def __contains__(self, image: Any) -> bool:
        """Checks if the given ``image`` is native for the
        :class:`~pyimagetest.backends.backend.ImageBackend`

        Args:
            image: Image to be checked
        """
        return isinstance(image, self.native_image_type)

    @abstractmethod
    def import_image(self, file: str) -> Any:
        """Imports an image form ``file`` into the
        :attr:`~pyimagetest.backends.backend.ImageBackend.native_image_type` of
        the backend.

        Args:
            file: Path to the file that should be imported.
        """
        pass

    @abstractmethod
    def export_image(self, image: Any) -> np.ndarray:
        """Exports an image of the
        :attr:`~pyimagetest.backends.backend.ImageBackend.native_image_type`
        into a
        `numpy.ndarray <https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.html>`_
        The output is of ``shape == (height, width, channels)`` and of
        ``dtype == numpy.float32``.

        Args:
            image: Image to be exported.
        """
        pass



from typing import Union, Optional, Collection, Dict
from importlib import import_module
from collections import OrderedDict
from .backend import ImageBackend


class BackendMeta:
    def __init__(
        self,
        requires: Union[Collection[str], str],
        module: str,
        name: Optional[str] = None,
        class_name: Optional[str] = None,
    ):
        if isinstance(requires, str):
            requires = (requires,)
        self.requires = requires

        self.module = module

        if name is None:
            name = module
        self.name = name

        if class_name is None:
            class_name = f"{name}Backend"
        self.class_name = class_name


BUILTIN_IMAGE_BACKENDS_META = (
    BackendMeta(requires="imageio", module="imageio"),
    BackendMeta(requires="PIL", module="Pillow", name="PIL"),
    BackendMeta(requires=("torch", "PIL", "torchvision"), module="torchvision"),
)


def builtin_image_backends() -> Dict[str, ImageBackend]:
    """Returns all available builtin image backends.

    ReturnType:
        OrderedDict[str, ImageBackend]
    """
    available_backends = OrderedDict(())
    for meta in BUILTIN_IMAGE_BACKENDS_META:
        try:
            for package in meta.requires:
                exec(f"import {package}")
        except ImportError:
            pass
        else:
            backend_class = getattr(
                import_module(f"pyimagetest.backends.{meta.module}"), meta.class_name
            )
            available_backends[meta.name] = backend_class()

    return available_backends



from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

about = {}
with open(path.join(here, "pyimagetest", "__about__.py"), "r") as fh:
    exec(fh.read(), about)

with open(path.join(here, "README.md"), "r") as fh:
    long_description = fh.read()

install_requires = ("numpy",)

builtin_backends_requires = (
    "imageio",
    "pillow",
    "torchvision",
)

type_check_requires = (
    "mypy",
    # TODO: move to a released version if available
    # FIXME: fix this for PyPI upload
    # "numpy-stubs@https://github.com/numpy/numpy-stubs/archive/f3c6315738489983f5f37e1477ac68373d71b470.zip",
)

test_requires = ("pytest",)

doc_requires = (
    "sphinx",
    "sphinx_autodoc_typehints",
    "sphinx_rtd_theme",
)

dev_requires = (
    *builtin_backends_requires,
    *type_check_requires,
    *test_requires,
    *doc_requires,
)

extras_require = {
    "builtin_backends": builtin_backends_requires,
    "type_check": type_check_requires,
    "test": test_requires,
    "doc": doc_requires,
    "dev": dev_requires,
}

classifiers = (
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Python :: 3",
    "Topic :: Software Development :: Testing :: Unit",
)

setup(
    name=about["__name__"],
    description=about["__description__"],
    version=about["__version__"],
    url=about["__url__"],
    license=about["__license__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(where=here, exclude=("third_party_stubs", "test", "docs")),
    install_requires=install_requires,
    extras_require=extras_require,
    python_requires=">=3.6",
    classifiers=classifiers,
)



from typing import Type
import numpy as np
import torch
from PIL import Image
from torchvision.transforms import functional as F
from .backend import ImageBackend

__all__ = ["torchvisionBackend"]


class torchvisionBackend(ImageBackend):
    """:class:`~pyimagetest.backends.backend.ImageBackend` for the
    `torchvision package <https://pytorch.org/docs/stable/torchvision/index.html>`_ .
    """

    @property
    def native_image_type(self) -> Type[torch.Tensor]:
        return torch.Tensor

    def import_image(self, file: str) -> torch.Tensor:
        return F.to_tensor(Image.open(file))

    def export_image(self, image: torch.Tensor) -> np.ndarray:
        return image.detach().cpu().permute((1, 2, 0)).numpy()



from typing import Type
import numpy as np
import imageio
from .backend import ImageBackend

__all__ = ["imageioBackend"]


class imageioBackend(ImageBackend):
    """:class:`~pyimagetest.backends.backend.ImageBackend` for the
    `imageio package <https://imageio.github.io/>`_ .
    """

    @property
    def native_image_type(self) -> Type[np.ndarray]:
        return np.ndarray

    def import_image(self, file: str) -> np.ndarray:
        return imageio.imread(file)

    def export_image(self, image: np.ndarray) -> np.ndarray:
        return image.astype(np.float32) / 255.0
